@@ -9,7 +9,7 @@ test_that("can work for correlation r2 between DS and GT", {
   res <- vcfcomp(imputedvcf, imputedvcf, bins = c(0,1),
                  by.sample = TRUE, samples = samples, stats = "all",
                  setid = TRUE)
-  expect_equal(as.numeric(res$r2[1,]), c(15, 6, 1), tolerance=1e-6)
+  expect_equal(as.numeric(unlist(res$r2[1,])), c(15, 6, 1, 1), tolerance=1e-6)
   expect_identical(unlist(res$f1[[3]]), rep(1,2))
   expect_identical(unlist(res$nrc[[3]]), rep(1,2))
 })
